@@ -87,7 +87,7 @@ def get_user_id(session: aiohttp.ClientSession) -> str:
         raise _exception.ParseError("Could not find user id", data=session.cookie_jar._cookies)
     if rtn is None:
         raise _exception.ParseError("Could not find user id", data=session.cookie_jar._cookies)
-    return str(rtn.value)
+    return rtn if isinstance(rtn, str) else str(rtn.value)
 
 
 def session_factory() -> aiohttp.ClientSession:
@@ -112,7 +112,7 @@ def find_form_request(html: str):
 
     form = soup.form
     if not form:
-        raise _exception.ParseError("Could not find form to submit", data=soup)
+        raise _exception.ParseError("Could not find form to submit", data=html)
 
     url = form.get("action")
     if not url:
@@ -141,6 +141,7 @@ async def two_factor_helper(session: aiohttp.ClientSession, r: aiohttp.ClientRes
         data["approvals_code"] = await on_2fa_callback()
         log.info("Submitting 2FA code")
         r = await session.post(url, data=data, allow_redirects=False)
+        log.debug("2FA location: %s", r.headers.get("Location"))
         url, data = find_form_request(await r.text())
 
     # TODO: Can be missing if checkup flow was done on another device in the meantime?
@@ -148,10 +149,12 @@ async def two_factor_helper(session: aiohttp.ClientSession, r: aiohttp.ClientRes
         data["name_action_selected"] = "save_device"
         log.info("Saving browser")
         r = await session.post(url, data=data, allow_redirects=False)
+        log.debug("2FA location: %s", r.headers.get("Location"))
         url, data = find_form_request(await r.text())
 
     log.info("Starting Facebook checkup flow")
     r = await session.post(url, data=data, allow_redirects=False)
+    log.debug("2FA location: %s", r.headers.get("Location"))
 
     url, data = find_form_request(await r.text())
     if "submit[This was me]" not in data or "submit[This wasn't me]" not in data:
@@ -159,15 +162,18 @@ async def two_factor_helper(session: aiohttp.ClientSession, r: aiohttp.ClientRes
     data["submit[This was me]"] = "[any value]"
     del data["submit[This wasn't me]"]
     log.info("Verifying login attempt")
+
     r = await session.post(url, data=data, allow_redirects=False)
+    log.debug("2FA location: %s", r.headers.get("Location"))
 
     url, data = find_form_request(await r.text())
     if "name_action_selected" not in data:
         raise _exception.ParseError("Could not fill out form properly (3)", data=data)
     data["name_action_selected"] = "save_device"
     log.info("Saving device again")
-    r = await session.post(url, data=data, allow_redirects=False)
 
+    r = await session.post(url, data=data, allow_redirects=False)
+    log.debug("2FA location: %s", r.headers.get("Location"))
     return r.headers.get("Location")
 
 
@@ -299,15 +305,22 @@ class Session:
                 raise _exception.NotLoggedIn(
                     "2FA code required! Please supply `on_2fa_callback` to .login"
                 )
-            # Get a facebook.com url that handles the 2FA flow
+            # Get a facebook.com/checkpoint/start url that handles the 2FA flow
             # This probably works differently for Messenger-only accounts
             url = _util.get_url_parameter(url, "next")
-            # Explicitly allow redirects
-            r = await session.get(url, allow_redirects=True)
+            if not url.startswith("https://www.facebook.com/checkpoint/start/"):
+                raise _exception.ParseError("Failed 2fa flow (1)", data=url)
+
+            r = await session.get(url, allow_redirects=False)
+            url = r.headers.get("Location")
+            if not url or not url.startswith("https://www.facebook.com/checkpoint/"):
+                raise _exception.ParseError("Failed 2fa flow (2)", data=url)
+
+            r = await session.get(url, allow_redirects=False)
             url = await two_factor_helper(session, r, on_2fa_callback)
 
             if not url.startswith("https://www.messenger.com/login/auth_token/"):
-                raise _exception.ParseError("Failed 2fa flow", data=url)
+                raise _exception.ParseError("Failed 2fa flow (3)", data=url)
 
             r = await session.get(url, allow_redirects=False)
             url = r.headers.get("Location")
